###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    This test is to establish that the core chemistry packages in IDAES solve
    a simple water dissociation problem and return the correct pH value, as well
    as excerising the implementation of the ENRTL model within that same context.

    This test also includes checking the pH value of a typical acid and checking
    the calculated activity coefficients for a relatively dilute system of ions.
"""
# Importing testing libraries
import pytest

# Importing the object for units from pyomo
from pyomo.environ import units as pyunits

# Imports from idaes core
from idaes.core import AqueousPhase
from idaes.core.components import Solvent, Solute, Cation, Anion
from idaes.core.phases import PhaseType as PT

# Imports from idaes generic models
import idaes.generic_models.properties.core.pure.Perrys as Perrys
from idaes.generic_models.properties.core.pure.electrolyte import relative_permittivity_constant
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.eos.enrtl import ENRTL
from idaes.generic_models.properties.core.eos.enrtl_reference_states import Unsymmetric

# Importing the enum for concentration unit basis used in the 'get_concentration_term' function
from idaes.generic_models.properties.core.generic.generic_reaction import ConcentrationForm

# Import the object/function for heat of reaction
from idaes.generic_models.properties.core.reactions.dh_rxn import constant_dh_rxn

# Import safe log power law equation
from idaes.generic_models.properties.core.reactions.equilibrium_forms import log_power_law_equil

# Import built-in van't Hoff function
from idaes.generic_models.properties.core.reactions.equilibrium_constant import van_t_hoff

# Import specific pyomo objects
from pyomo.environ import (ConcreteModel,
                           SolverStatus,
                           TerminationCondition,
                           value,
                           Suffix)

# Import the scaling methods
from idaes.core.util import scaling as iscale

# Import pyomo methods to check the system units
from pyomo.util.check_units import assert_units_consistent

# Import idaes methods to check the model during construction
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables)

# Import the idaes objects for Generic Properties and Reactions
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock)
from idaes.generic_models.properties.core.generic.generic_reaction import (
        GenericReactionParameterBlock)

# Import the idaes object for the EquilibriumReactor unit model
from idaes.generic_models.unit_models.equilibrium_reactor import EquilibriumReactor

# Import the core idaes objects for Flowsheets and types of balances
from idaes.core import FlowsheetBlock

# Import log10 function from pyomo
from pyomo.environ import log10

import idaes.logger as idaeslog

__author__ = "Austin Ladshaw"

# Configuration dictionary
water_thermo_config = {
    "components": {
        'H2O': {"type": Solvent,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              "relative_permittivity_liq_comp": relative_permittivity_constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (18.0153, pyunits.g/pyunits.mol),
                    "relative_permittivity_liq_comp": 78.54,
                    "pressure_crit": (220.64E5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    # Comes from Perry's Handbook:  p. 2-98
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-285.830, pyunits.kJ/pyunits.mol),
                    "enth_mol_form_vap_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    # Comes from Perry's Handbook:  p. 2-174
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "cp_mol_ig_comp_coeff": {
                        'A': (30.09200, pyunits.J/pyunits.mol/pyunits.K),
                        'B': (6.832514, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-1),
                        'C': (6.793435, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-2),
                        'D': (-2.534480, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-3),
                        'E': (0.082139, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**2),
                        'F': (-250.8810, pyunits.kJ/pyunits.mol),
                        'G': (223.3967, pyunits.J/pyunits.mol/pyunits.K),
                        'H': (0, pyunits.kJ/pyunits.mol)},
                    "entr_mol_form_liq_comp_ref": (69.95, pyunits.J/pyunits.K/pyunits.mol),
                    "pressure_sat_comp_coeff": {
                        'A': (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                        'B': (1435.264, pyunits.K),
                        'C': (-64.848, pyunits.K)}
                                },
                    # End parameter_data
                    },
        'H_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (1.00784, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (-10.75, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OH_-': {"type": Anion,
                "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (17.008, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (-10.75, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    }
              },
              # End Component list
        "phases":  {'Liq': {"type": AqueousPhase,
                            "equation_of_state": ENRTL,
                            "equation_of_state_options": {
                                    "reference_state": Unsymmetric}
                            },
                    },

        "state_definition": FTPx,
        "state_bounds": {"flow_mol": (0, 50, 100),
                         "temperature": (273.15, 300, 650),
                         "pressure": (5e4, 1e5, 1e6)
                     },

        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

        # Inherent reactions
        "inherent_reactions": {
            "H2O_Kw": {
                    "stoichiometry": {("Liq", "H2O"): -1,
                                     ("Liq", "H_+"): 1,
                                     ("Liq", "OH_-"): 1},
                   "heat_of_reaction": constant_dh_rxn,
                   "equilibrium_constant": van_t_hoff,
                   "equilibrium_form": log_power_law_equil,
                   "concentration_form": ConcentrationForm.activity,
                   "parameter_data": {
                       #NOTE: The k value on the activity basis is UNITLESS
                       #        based on a standard molar concentration of 1 mol/L
                       #        HOWEVER, the typical Kw dissociation constant of
                       #        1e-14 is defined on a molar basis. Thus, we must
                       #        divide by the total (~55.2 M) concentration raised to the
                       #        net reaction order (i.e., 2 in this case).
                       "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                       "k_eq_ref": (10**-14/55.2**2,pyunits.dimensionless),
                       "T_eq_ref": (298, pyunits.K),

                       # By default, reaction orders follow stoichiometry
                       #    manually set reaction order here to override
                       "reaction_order": {("Liq", "H2O"): 0,
                                        ("Liq", "H_+"): 1,
                                        ("Liq", "OH_-"): 1}
                        }
                        # End parameter_data
                   }
                   # End R1
             }
             # End equilibrium_reactions
    }
    # End thermo_config definition

# Define the reaction_config for water dissociation
# This config is REQUIRED to use EquilibriumReactor even if we have no equilibrium reactions
reaction_config = {
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "equilibrium_reactions": {
        "dummy": {
                "stoichiometry": {},
                "equilibrium_form": log_power_law_equil,
               }
         }
         # End equilibrium_reactions
    }
    # End reaction_config definition

# Get default solver for testing
solver = get_solver()

# Start test class
class TestENRTLwater():
    @pytest.fixture(scope="class")
    def water_model(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(default={"dynamic": False})
        model.fs.thermo_params = GenericParameterBlock(default=water_thermo_config)
        model.fs.rxn_params = GenericReactionParameterBlock(
                default={"property_package": model.fs.thermo_params, **reaction_config})
        model.fs.unit = EquilibriumReactor(default={
                "property_package": model.fs.thermo_params,
                "reaction_package": model.fs.rxn_params,
                "has_rate_reactions": False,
                "has_equilibrium_reactions": False,
                "has_heat_transfer": False,
                "has_heat_of_reaction": False,
                "has_pressure_change": False})

        #NOTE: ENRTL model cannot initialize if the inlet values are 0
        zero = 1e-20
        model.fs.unit.inlet.mole_frac_comp[0, "H_+"].fix( zero )
        model.fs.unit.inlet.mole_frac_comp[0, "OH_-"].fix( zero )
        model.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix( 1.-2*zero )
        model.fs.unit.inlet.pressure.fix(101325.0)
        model.fs.unit.inlet.temperature.fix(298.)
        model.fs.unit.inlet.flow_mol.fix(10)

        return model

    @pytest.mark.unit
    def test_build_water(self, water_model):
        model = water_model

        assert hasattr(model.fs.thermo_params, 'component_list')
        assert len(model.fs.thermo_params.component_list) == 3
        assert 'H2O' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.H2O, Solvent)
        assert 'H_+' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component('H_+'), Cation)
        assert 'OH_-' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component('OH_-'), Anion)

    @pytest.mark.unit
    def test_units_water(self, water_model):
        model = water_model
        assert_units_consistent(model)

    @pytest.mark.unit
    def test_dof_water(self, water_model):
        model = water_model
        assert (degrees_of_freedom(model) == 0)

    @pytest.mark.component
    def test_scaling_water(self, water_model):
        model = water_model

        for i in model.fs.unit.control_volume.inherent_reaction_extent_index:
            scale = value(model.fs.unit.control_volume.properties_out[0.0].k_eq[i[1]].expr)
            iscale.set_scaling_factor(model.fs.unit.control_volume.inherent_reaction_extent[0.0,i[1]], 10/scale)
            iscale.constraint_scaling_transform(model.fs.unit.control_volume.properties_out[0.0].
                    inherent_equilibrium_constraint[i[1]], 0.1)

        # Next, try adding scaling for species
        min = 1e-3
        for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
            # i[0] = phase, i[1] = species
            if model.fs.unit.inlet.mole_frac_comp[0, i[1]].value > min:
                scale = model.fs.unit.inlet.mole_frac_comp[0, i[1]].value
            else:
                scale = min
            iscale.set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]], 10/scale)
            iscale.set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i], 10/scale)
            iscale.set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i], 10/scale)
            iscale.constraint_scaling_transform(
                model.fs.unit.control_volume.properties_out[0.0].component_flow_balances[i[1]], 10/scale)
            iscale.constraint_scaling_transform(model.fs.unit.control_volume.material_balances[0.0,i[1]], 10/scale)

        iscale.calculate_scaling_factors(model.fs.unit)

        assert hasattr(model.fs.unit.control_volume, 'scaling_factor')
        assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)

        assert hasattr(model.fs.unit.control_volume.properties_out[0.0], 'scaling_factor')
        assert isinstance(model.fs.unit.control_volume.properties_out[0.0].scaling_factor, Suffix)

        assert hasattr(model.fs.unit.control_volume.properties_in[0.0], 'scaling_factor')
        assert isinstance(model.fs.unit.control_volume.properties_in[0.0].scaling_factor, Suffix)

    @pytest.mark.component
    def test_initialize_solver_water(self, water_model):
        model = water_model
        model.fs.unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_solve_water(self, water_model):
        model = water_model
        solver.options['max_iter'] = 200
        solver.options['halt_on_ampl_error'] = 'yes'
        results = solver.solve(model, tee=True, symbolic_solver_labels=True)
        print(results.solver.termination_condition)
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_solution_water(self, water_model):
        model = water_model

        pH = -log10(value(model.fs.unit.control_volume.properties_out[0.0].act_phase_comp["Liq","H_+"])*55.2)
        pOH = -log10(value(model.fs.unit.control_volume.properties_out[0.0].act_phase_comp["Liq","OH_-"])*55.2)

        assert pytest.approx(7.0001611, rel=1e-4) == pH
        assert pytest.approx(7.0001611, rel=1e-4) == pOH

# Configuration dictionary
carbonic_thermo_config = {
    "components": {
        'H2O': {"type": Solvent,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              "relative_permittivity_liq_comp": relative_permittivity_constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (18.0153, pyunits.g/pyunits.mol),
                    "relative_permittivity_liq_comp": 78.54,
                    "pressure_crit": (220.64E5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    # Comes from Perry's Handbook:  p. 2-98
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-285.830, pyunits.kJ/pyunits.mol),
                    "enth_mol_form_vap_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    # Comes from Perry's Handbook:  p. 2-174
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "cp_mol_ig_comp_coeff": {
                        'A': (30.09200, pyunits.J/pyunits.mol/pyunits.K),
                        'B': (6.832514, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-1),
                        'C': (6.793435, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-2),
                        'D': (-2.534480, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**-3),
                        'E': (0.082139, pyunits.J*pyunits.mol**-1*pyunits.K**-1*pyunits.kiloK**2),
                        'F': (-250.8810, pyunits.kJ/pyunits.mol),
                        'G': (223.3967, pyunits.J/pyunits.mol/pyunits.K),
                        'H': (0, pyunits.kJ/pyunits.mol)},
                    "entr_mol_form_liq_comp_ref": (69.95, pyunits.J/pyunits.K/pyunits.mol),
                    "pressure_sat_comp_coeff": {
                        'A': (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                        'B': (1435.264, pyunits.K),
                        'C': (-64.848, pyunits.K)}
                                },
                    # End parameter_data
                    },
        'H_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (1.00784, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (-10.75, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OH_-': {"type": Anion,
                "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (17.008, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.459, pyunits.kmol*pyunits.m**-3),
                        '2': (0.30542, pyunits.dimensionless),
                        '3': (647.13, pyunits.K),
                        '4': (0.081, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-230.000, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (2.7637E5, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (-2.0901E3, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (8.125, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (-1.4116E-2, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (9.3701E-6, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (-10.75, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H2CO3': {"type": Solute, "valid_phase_types": PT.aqueousPhase,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.4495, pyunits.kmol*pyunits.m**-3),
                        '2': (0.427, pyunits.dimensionless),
                        '3': (429.69, pyunits.K),
                        '4': (0.259, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-699.7, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (135749.9, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (187, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'HCO3_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (61.0168, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.4495, pyunits.kmol*pyunits.m**-3),
                        '2': (0.427, pyunits.dimensionless),
                        '3': (429.69, pyunits.K),
                        '4': (0.259, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-692, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (135749.9, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (91.2, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'CO3_2-': {"type": Anion, "charge": -2,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (60.01, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.4495, pyunits.kmol*pyunits.m**-3),
                        '2': (0.427, pyunits.dimensionless),
                        '3': (429.69, pyunits.K),
                        '4': (0.259, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-677.1, pyunits.J/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (135749.9, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (-56.9, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Na_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Perrys,
              "enth_mol_liq_comp": Perrys,
              "cp_mol_liq_comp": Perrys,
              "entr_mol_liq_comp": Perrys,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (22.989769, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": {
                        '1': (5.252, pyunits.kmol*pyunits.m**-3),
                        '2': (0.347, pyunits.dimensionless),
                        '3': (1595.8, pyunits.K),
                        '4': (0.6598, pyunits.dimensionless)},
                    "enth_mol_form_liq_comp_ref": (-240.1, pyunits.J/pyunits.mol),
                    "cp_mol_liq_comp_coeff": {
                        '1': (167039, pyunits.J/pyunits.kmol/pyunits.K),
                        '2': (0, pyunits.J/pyunits.kmol/pyunits.K**2),
                        '3': (0, pyunits.J/pyunits.kmol/pyunits.K**3),
                        '4': (0, pyunits.J/pyunits.kmol/pyunits.K**4),
                        '5': (0, pyunits.J/pyunits.kmol/pyunits.K**5)},
                    "entr_mol_form_liq_comp_ref": (59, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    }
              },
              # End Component list
        "phases":  {'Liq': {"type": AqueousPhase,
                            "equation_of_state": ENRTL,
                            "equation_of_state_options": {
                                    "reference_state": Unsymmetric}
                            },
                    },

        "state_definition": FTPx,
        "state_bounds": {"flow_mol": (0, 50, 100),
                         "temperature": (273.15, 300, 650),
                         "pressure": (5e4, 1e5, 1e6)
                     },

        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

        # Inherent reactions
        "inherent_reactions": {
            "H2O_Kw": {
                    "stoichiometry": {("Liq", "H2O"): -1,
                                     ("Liq", "H_+"): 1,
                                     ("Liq", "OH_-"): 1},
                   "heat_of_reaction": constant_dh_rxn,
                   "equilibrium_constant": van_t_hoff,
                   "equilibrium_form": log_power_law_equil,
                   "concentration_form": ConcentrationForm.activity,
                   "parameter_data": {
                       #NOTE: The k value on the activity basis is UNITLESS
                       #        based on a standard molar concentration of 1 mol/L
                       #        HOWEVER, the typical Kw dissociation constant of
                       #        1e-14 is defined on a molar basis. Thus, we must
                       #        divide by the total (~55.2 M) concentration raised to the
                       #        net reaction order (i.e., 2 in this case).
                       "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                       "k_eq_ref": (10**-14/55.2**2, pyunits.dimensionless),
                       "T_eq_ref": (298, pyunits.K),

                       # By default, reaction orders follow stoichiometry
                       #    manually set reaction order here to override
                       "reaction_order": {("Liq", "H2O"): 0,
                                        ("Liq", "H_+"): 1,
                                        ("Liq", "OH_-"): 1}
                        }
                        # End parameter_data
                   },
                   # End R1
            "H2CO3_Ka1": {
                    "stoichiometry": {("Liq", "H2CO3"): -1,
                                     ("Liq", "H_+"): 1,
                                     ("Liq", "HCO3_-"): 1},
                   "heat_of_reaction": constant_dh_rxn,
                   "equilibrium_constant": van_t_hoff,
                   "equilibrium_form": log_power_law_equil,
                   "concentration_form": ConcentrationForm.activity,
                   "parameter_data": {
                       #NOTE: The k value on the activity basis is UNITLESS
                       #        based on a standard molar concentration of 1 mol/L
                       #        HOWEVER, the typical Ka1 dissociation constant of
                       #        10**-6.33 is defined on a molar basis. Thus, we must
                       #        divide by the total (~55.2 M) concentration raised to the
                       #        net reaction order (i.e., 1 in this case).
                       "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                       "k_eq_ref": (10**-6.33/55.2,pyunits.dimensionless),
                       "T_eq_ref": (300, pyunits.K),

                       # By default, reaction orders follow stoichiometry
                       #    manually set reaction order here to override
                       "reaction_order": {("Liq", "H2CO3"): -1,
                                        ("Liq", "H_+"): 1,
                                        ("Liq", "HCO3_-"): 1}
                        }
                        # End parameter_data
                   },
                   # End R2
            "H2CO3_Ka2": {
                    "stoichiometry": {("Liq", "HCO3_-"): -1,
                                     ("Liq", "H_+"): 1,
                                     ("Liq", "CO3_2-"): 1},
                   "heat_of_reaction": constant_dh_rxn,
                   "equilibrium_constant": van_t_hoff,
                   "equilibrium_form": log_power_law_equil,
                   "concentration_form": ConcentrationForm.activity,
                   "parameter_data": {
                       #NOTE: The k value on the activity basis is UNITLESS
                       #        based on a standard molar concentration of 1 mol/L
                       #        HOWEVER, the typical Ka1 dissociation constant of
                       #        10**-10.35 is defined on a molar basis. Thus, we must
                       #        divide by the total (~55.2 M) concentration raised to the
                       #        net reaction order (i.e., 1 in this case).
                       "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                       "k_eq_ref": (10**-10.35/55.2,pyunits.dimensionless),
                       "T_eq_ref": (300, pyunits.K),

                       # By default, reaction orders follow stoichiometry
                       #    manually set reaction order here to override
                       "reaction_order": {("Liq", "HCO3_-"): -1,
                                        ("Liq", "H_+"): 1,
                                        ("Liq", "CO3_2-"): 1}
                        }
                        # End parameter_data
                   }
                   # End R3
             }
             # End equilibrium_reactions
    }
    # End thermo_config definition

# Start test class
class TestENRTLcarbonicAcid():
    @pytest.fixture(scope="class")
    def carbonic_acid_model(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(default={"dynamic": False})
        model.fs.thermo_params = GenericParameterBlock(default=carbonic_thermo_config)
        model.fs.rxn_params = GenericReactionParameterBlock(
                default={"property_package": model.fs.thermo_params, **reaction_config})
        model.fs.unit = EquilibriumReactor(default={
                "property_package": model.fs.thermo_params,
                "reaction_package": model.fs.rxn_params,
                "has_rate_reactions": False,
                "has_equilibrium_reactions": False,
                "has_heat_transfer": False,
                "has_heat_of_reaction": False,
                "has_pressure_change": False})

        #NOTE: ENRTL model cannot initialize if the inlet values are 0
        zero = 1e-20
        acid = 0.00206/(55.2+0.00206)
        model.fs.unit.inlet.mole_frac_comp[0, "H_+"].fix( zero )
        model.fs.unit.inlet.mole_frac_comp[0, "OH_-"].fix( zero )

        # Added as conjugate base form
        model.fs.unit.inlet.mole_frac_comp[0, "CO3_2-"].fix( zero )
        model.fs.unit.inlet.mole_frac_comp[0, "HCO3_-"].fix( acid )
        model.fs.unit.inlet.mole_frac_comp[0, "H2CO3"].fix( zero )
        model.fs.unit.inlet.mole_frac_comp[0, "Na_+"].fix( acid )

        model.fs.unit.inlet.mole_frac_comp[0, "H2O"].fix( 1.-4*zero-acid- \
                        value(model.fs.unit.inlet.mole_frac_comp[0, "Na_+"]) )
        model.fs.unit.inlet.pressure.fix(101325.0)
        model.fs.unit.inlet.temperature.fix(298.)
        model.fs.unit.inlet.flow_mol.fix(10)

        return model

    @pytest.mark.unit
    def test_build_carbonic_acid(self, carbonic_acid_model):
        model = carbonic_acid_model

        assert hasattr(model.fs.thermo_params, 'component_list')
        assert len(model.fs.thermo_params.component_list) == 7
        assert 'H2O' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.H2O, Solvent)
        assert 'H_+' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component('H_+'), Cation)
        assert 'OH_-' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component('OH_-'), Anion)

        assert 'Na_+' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component('Na_+'), Cation)
        assert 'HCO3_-' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component('HCO3_-'), Anion)
        assert 'CO3_2-' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.component('CO3_2-'), Anion)
        assert 'H2CO3' in model.fs.thermo_params.component_list
        assert isinstance(model.fs.thermo_params.H2CO3, Solute)

    #NOTE: This test takes ~8s and I have no idea why...
    @pytest.mark.unit
    def test_units_carbonic_acid(self, carbonic_acid_model):
        model = carbonic_acid_model
        assert_units_consistent(model)

    @pytest.mark.unit
    def test_dof_carbonic_acid(self, carbonic_acid_model):
        model = carbonic_acid_model
        assert (degrees_of_freedom(model) == 0)

    @pytest.mark.component
    def test_scaling_carbonic_acid(self, carbonic_acid_model):
        model = carbonic_acid_model

        for i in model.fs.unit.control_volume.inherent_reaction_extent_index:
            scale = value(model.fs.unit.control_volume.properties_out[0.0].k_eq[i[1]].expr)
            iscale.set_scaling_factor(model.fs.unit.control_volume.inherent_reaction_extent[0.0,i[1]], 10/scale)
            iscale.constraint_scaling_transform(model.fs.unit.control_volume.properties_out[0.0].
                    inherent_equilibrium_constraint[i[1]], 0.1)

        # Next, try adding scaling for species
        min = 1e-3
        for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
            # i[0] = phase, i[1] = species
            if model.fs.unit.inlet.mole_frac_comp[0, i[1]].value > min:
                scale = model.fs.unit.inlet.mole_frac_comp[0, i[1]].value
            else:
                scale = min
            iscale.set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]], 10/scale)
            iscale.set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i], 10/scale)
            iscale.set_scaling_factor(model.fs.unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i], 10/scale)
            iscale.constraint_scaling_transform(
                model.fs.unit.control_volume.properties_out[0.0].component_flow_balances[i[1]], 10/scale)
            iscale.constraint_scaling_transform(model.fs.unit.control_volume.material_balances[0.0,i[1]], 10/scale)

        iscale.calculate_scaling_factors(model.fs.unit)

        assert hasattr(model.fs.unit.control_volume, 'scaling_factor')
        assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)

        assert hasattr(model.fs.unit.control_volume.properties_out[0.0], 'scaling_factor')
        assert isinstance(model.fs.unit.control_volume.properties_out[0.0].scaling_factor, Suffix)

        assert hasattr(model.fs.unit.control_volume.properties_in[0.0], 'scaling_factor')
        assert isinstance(model.fs.unit.control_volume.properties_in[0.0].scaling_factor, Suffix)

    @pytest.mark.component
    def test_initialize_solver_carbonic_acid(self, carbonic_acid_model):
        model = carbonic_acid_model
        solver.options['max_iter'] = 100
        model.fs.unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_solve_carbonic_acid(self, carbonic_acid_model):
        model = carbonic_acid_model
        results = solver.solve(model, tee=True)
        print(results.solver.termination_condition)
        assert results.solver.termination_condition == TerminationCondition.optimal
        assert results.solver.status == SolverStatus.ok

    @pytest.mark.component
    def test_solution_carbonic_acid(self, carbonic_acid_model):
        model = carbonic_acid_model

        pH = -log10(value(model.fs.unit.control_volume.properties_out[0.0].act_phase_comp["Liq","H_+"])*55.2)
        pOH = -log10(value(model.fs.unit.control_volume.properties_out[0.0].act_phase_comp["Liq","OH_-"])*55.2)

        assert pytest.approx(8.28, rel=1e-2) == pH
        assert pytest.approx(5.72, rel=1e-2) == pOH
        assert pytest.approx(14.00, rel=1e-2) == pH + pOH

        gamma = {}
        for index in model.fs.unit.control_volume.properties_out[0.0].conc_mol_phase_comp:
            gamma[index] = value(model.fs.unit.control_volume.properties_out[0.0].act_phase_comp[index])/value(model.fs.unit.outlet.mole_frac_comp[0,index[1]])

        assert pytest.approx(0.95075, rel=1e-5) == gamma[('Liq', 'OH_-')]
        assert pytest.approx(0.95075, rel=1e-5) == gamma[('Liq', 'H_+')]
        assert pytest.approx(0.95075, rel=1e-5) == gamma[('Liq', 'Na_+')]
        assert pytest.approx(0.95075, rel=1e-5) == gamma[('Liq', 'HCO3_-')]
        assert pytest.approx(0.81710, rel=1e-5) == gamma[('Liq', 'CO3_2-')]
        assert pytest.approx(1.00000, rel=1e-5) == gamma[('Liq', 'H2O')]
        assert pytest.approx(1.00000, rel=1e-5) == gamma[('Liq', 'H2CO3')]

        total_acid = value(model.fs.unit.control_volume.properties_out[0.0].conc_mol_phase_comp["Liq","H2CO3"])/1000
        total_acid += value(model.fs.unit.control_volume.properties_out[0.0].conc_mol_phase_comp["Liq","HCO3_-"])/1000
        total_acid += value(model.fs.unit.control_volume.properties_out[0.0].conc_mol_phase_comp["Liq","CO3_2-"])/1000

        assert pytest.approx(0.002061178349769601, rel=1e-5) == total_acid



"""
Test Jupyter notebooks in the docs.
"""
# stdlib
import logging
from pathlib import Path
import time

# deps
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor, CellExecutionError
import pytest

# package
from watertap.edb import ElectrolyteDB

# Logging setup

_log = logging.getLogger(__name__)

# Constants

JUPYTER_NB_VERSION = 4
EXECUTE_TIMEOUT_SEC = 600

# Exceptions


class NotebookError(Exception):
    pass


# Utility
# =======


def run_notebook(path: Path):
    # parse
    _log.debug(f"begin: parsing '{path}'")
    with path.open("r", encoding="utf-8") as f:
        nb = nbformat.read(f, as_version=JUPYTER_NB_VERSION)
    _log.debug(f"end: parsing '{path}'")

    # execute
    pp = ExecutePreprocessor(timeout=EXECUTE_TIMEOUT_SEC)
    _log.debug(f"begin: executing '{path}'")
    t0 = time.time()
    try:
        metadata = {"metadata": {"path": path.parent}}
        pp.preprocess(nb, metadata)
    except (CellExecutionError, NameError) as err:
        _log.warning(f"Failed: {path} :: {err}")
        raise NotebookError(f"execution error for '{path}': {err}")
    except TimeoutError as err:
        _log.warning(f"Timeout: {path} :: {err}")
        dur, timeout = time.time() - t0, EXECUTE_TIMEOUT_SEC
        raise NotebookError(f"timeout for '{path}': {dur}s > {timeout}s")
    dur = time.time() - t0
    _log.debug(f"end: executing '{path}' duration={dur}s")


def find_notebooks(path: Path):
    return path.glob("**/*.ipynb")


# Tests
# =====


@pytest.mark.component
def test_edb_notebooks(docs_root, electrolytedb):
    for nb_path in find_notebooks(docs_root / "examples" / "edb"):
        if ".ipynb_checkpoints" in nb_path.parts:
            continue
        print(f"run notebook: {nb_path}")
        run_notebook(nb_path)




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains the base class for all zero order single inlet-two outlet
(SITO) unit models.
"""
from idaes.core import UnitModelBlockData, useDefault
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog
from idaes.core.util import get_solver
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import ConfigurationError

from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.environ import NonNegativeReals, Var, units as pyunits

# Some more inforation about this module
__author__ = "Andrew Lee"


class SITOBaseData(UnitModelBlockData):
    """
    Standard base class for single inlet-two outlet unit models.

    This class is intended to be used for creating derived model classes and
    cannot be instantiated by itself. When creating derived classes,
    developers must set the '_has_deltaP_treated` and `_has_deltaP_byproduct`
    attributes on the model before calling `super().build()`. These attributes
    determine whether the deltaP terms for the treated and byproduct streams
    are added to the model and included in the pressure constraints.
    """

    CONFIG = ConfigBlock()
    CONFIG.declare('dynamic', ConfigValue(
        domain=In([False]),
        default=False,
        description='Dynamic model flag - must be False',
        doc='''All zero-order models are steady-state only'''))
    CONFIG.declare('has_holdup', ConfigValue(
        default=False,
        domain=In([False]),
        description='Holdup construction flag - must be False',
        doc='''Zero order models do not include holdup'''))
    CONFIG.declare('property_package', ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description='Property package to use for control volume',
        doc='''Property parameter object used to define property  calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}'''))
    CONFIG.declare('property_package_args', ConfigBlock(
        implicit=True,
        description='Arguments to use for constructing property packages',
        doc='''A ConfigBlock with arguments to be passed to a property block(s)
        and used when constructing these, **default** - None.
        **Valid values:** {see property package for documentation.}'''))

    def build(self):
        super().build()

        # Check that derived class has implemented flags for pressure change
        if not hasattr(self, "_has_deltaP_treated"):
            raise NotImplementedError(
                f"{self.name} derived class has not been implemented "
                f"_has_deltaP_treated.")
        if not hasattr(self, "_has_deltaP_byproduct"):
            raise NotImplementedError(
                f"{self.name} derived class has not been implemented "
                f"_has_deltaP_byproduct.")

        # Check that property package meets requirements
        if self.config.property_package.phase_list != ["Liq"]:
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models only support property packages with a "
                f"single phase named 'Liq'.")
        if (not hasattr(self.config.property_package, "solvent_set") or
                self.config.property_package.solvent_set != ["H2O"]):
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models only support property packages which "
                f"include 'H2O' as the only Solvent.")
        if not hasattr(self.config.property_package, "solute_set"):
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models require property packages to declare all "
                f"dissolved species as Solutes.")
        if (len(self.config.property_package.solute_set) !=
                len(self.config.property_package.component_list)-1):
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models only support `H2O` as a solvent and all "
                f"other species as Solutes.")

        # Get units metadata
        units_meta = \
            self.config.property_package.get_metadata().get_derived_units

        # Create state blocks for inlet and outlets
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.properties_in = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties at inlet",
            default=tmp_dict)

        tmp_dict_2 = dict(**tmp_dict)
        tmp_dict_2["defined_state"] = False

        self.properties_treated = \
            self.config.property_package.build_state_block(
                self.flowsheet().time,
                doc="Material properties of treated water",
                default=tmp_dict_2)
        self.properties_byproduct = \
            self.config.property_package.build_state_block(
                self.flowsheet().time,
                doc="Material properties of byproduct stream",
                default=tmp_dict_2)

        # Create Ports
        self.add_port("inlet", self.properties_in, doc="Inlet port")
        self.add_port("treated",
                      self.properties_treated,
                      doc="Treated water outlet port")
        self.add_port("byproduct",
                      self.properties_byproduct,
                      doc="Byproduct outlet port")

        # Add performance variables
        self.recovery_vol = Var(
            self.flowsheet().time,
            initialize=0.8,
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            bounds=(1E-8, 1.0000001),
            doc='Volumetric recovery fraction of water in the treated stream')
        self.removal_mass_solute = Var(
            self.flowsheet().time,
            self.config.property_package.solute_set,
            domain=NonNegativeReals,
            initialize=0.01,
            units=pyunits.dimensionless,
            doc='Solute removal fraction on a mass basis')

        if self._has_deltaP_treated:
            self.deltaP_treated = Var(
                self.flowsheet().time,
                initialize=0,
                units=units_meta('pressure'),
                doc='Pressure change between inlet and treated outlet')
        if self._has_deltaP_byproduct:
            self.deltaP_byproduct = Var(
                self.flowsheet().time,
                initialize=0,
                units=units_meta('pressure'),
                doc='Pressure change between inlet and byproduct outlet')

        # Add performance constraints
        # Water recovery
        @self.Constraint(self.flowsheet().time, doc='Water recovery equation')
        def water_recovery_equation(b, t):
            return (b.recovery_vol[t] * b.properties_in[t].flow_vol ==
                    b.properties_treated[t].flow_vol)

        # Flow balance
        @self.Constraint(self.flowsheet().time, doc='Overall flow balance')
        def flow_balance(b, t):
            return (b.properties_in[t].flow_vol ==
                    b.properties_treated[t].flow_vol +
                    b.properties_byproduct[t].flow_vol)

        # Solute removal
        @self.Constraint(self.flowsheet().time,
                         self.config.property_package.solute_set,
                         doc='Solute removal equations')
        def solute_removal_equation(b, t, j):
            return (b.removal_mass_solute[t, j] *
                    b.properties_in[t].conc_mass_comp[j] ==
                    (1 - b.recovery_vol[t]) *
                    b.properties_byproduct[t].conc_mass_comp[j])

        # Solute concentration of treated stream
        @self.Constraint(self.flowsheet().time,
                         self.config.property_package.solute_set,
                         doc='Constraint for solute concentration in treated '
                         'stream.')
        def solute_treated_equation(b, t, j):
            return ((1 - b.removal_mass_solute[t, j]) *
                    b.properties_in[t].conc_mass_comp[j] ==
                    b.recovery_vol[t] *
                    b.properties_treated[t].conc_mass_comp[j])

        # Pressure drop
        @self.Constraint(self.flowsheet().time,
                         doc='Treated stream pressure equation')
        def treated_pressure_constraint(b, t):
            if self._has_deltaP_treated:
                dp = b.deltaP_treated[t]
            else:
                dp = 0
            return (b.properties_in[t].pressure + dp ==
                    b.properties_treated[t].pressure)

        @self.Constraint(self.flowsheet().time,
                         doc='Byproduct stream pressure equation')
        def byproduct_pressure_constraint(b, t):
            if self._has_deltaP_byproduct:
                dp = b.deltaP_byproduct[t]
            else:
                dp = 0
            return (b.properties_in[t].pressure + dp ==
                    b.properties_byproduct[t].pressure)

        # Temperature equality
        @self.Constraint(self.flowsheet().time,
                         doc='Treated stream temperature equality')
        def treated_temperature_equality(b, t):
            return (b.properties_in[t].temperature ==
                    b.properties_treated[t].temperature)

        @self.Constraint(self.flowsheet().time,
                         doc='Byproduct stream temperature equality')
        def byproduct_temperature_equality(b, t):
            return (b.properties_in[t].temperature ==
                    b.properties_byproduct[t].temperature)

    def initialize(blk, state_args=None, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        '''
        Initialization routine for single inlet-two outlet unit models.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default IDAES solver)

        Returns:
            None
        '''
        if optarg is None:
            optarg = {}

        # Set solver options
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        solver_obj = get_solver(solver, optarg)

        # Get initial guesses for inlet if none provided
        if state_args is None:
            state_args = {}
            state_dict = (
                blk.properties_in[
                    blk.flowsheet().time.first()]
                .define_port_members())

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        # ---------------------------------------------------------------------
        # Initialize control volume block
        flags = blk.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=True
        )
        blk.properties_treated.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=False
        )
        blk.properties_byproduct.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
            hold_state=False
        )

        init_log.info_high('Initialization Step 1 Complete.')

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = solver_obj.solve(blk, tee=slc.tee)

        init_log.info_high(
            "Initialization Step 2 {}.".format(idaeslog.condition(results))
        )

        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.properties_in.release_state(flags, outlvl)

        init_log.info('Initialization Complete: {}'
                      .format(idaeslog.condition(results)))

    def calculate_scaling_factors(self):
        # Get default scale factors and do calculations from base classes
        super().calculate_scaling_factors()

        for t, v in self.water_recovery_equation.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].flow_vol,
                    default=1,
                    warning=True,
                    hint=" for water recovery"))

        for t, v in self.flow_balance.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].flow_vol,
                    default=1,
                    warning=False))  # would just be a duplicate of above

        for (t, j), v in self.solute_removal_equation.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].flow_mass_comp[j],
                    default=1,
                    warning=True,
                    hint=" for solute removal"))

        for (t, j), v in self.solute_treated_equation.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].flow_mass_comp[j],
                    default=1,
                    warning=False))  # would just be a duplicate of above

        for t, v in self.treated_pressure_constraint.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].pressure,
                    default=1e-5,
                    warning=True,
                    hint=" for treated pressure constraint"))

        for t, v in self.byproduct_pressure_constraint.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].pressure,
                    default=1e-5,
                    warning=False))  # would just be a duplicate of above

        for t, v in self.treated_temperature_equality.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].temperature,
                    default=1e-2,
                    warning=True,
                    hint=" for treated temperature equality"))

        for t, v in self.byproduct_temperature_equality.items():
            iscale.constraint_scaling_transform(
                v, iscale.get_scaling_factor(
                    self.properties_in[t].temperature,
                    default=1e-2,
                    warning=False))  # would just be a duplicate of above



"""
Tests for documentation
"""
# stdlib
from pathlib import Path
import re
# third-party
import pytest

__authors__ = ("Dan Gunter",)


# root of documentation
docs_dir = Path(__file__).parent.parent.parent / "docs"
# generated by sphinx-apidoc command
apidoc = docs_dir / "apidoc"


def has_docs():
    # assume this file is under <root>/<package>/tests/ , docs in <root>/docs
    # check dir docs/apidoc dir
    if not apidoc.exists() or not apidoc.is_dir():
        return False
    # XXX: add test for Sphinx too?
    return True


def get_autodoc(pth):
    """Get the name, location, and options for all autodoc directives
    in the file at 'pth'.

    Yields:
        tuple of length 4: type of thing, name of it, filename, line number (from 1)
    """
    mode = "text"  # state machine with 2 states; other one is "autodoc"
    auto_what, auto_name, auto_options, auto_where = None, None, None, None
    with open(pth, encoding="utf-8") as f:
        for i, line in enumerate(f):
            if mode == "text":
                match = re.match(r'..\s+auto(\w+)::\s*([a-zA-Z][a-zaA-Z_.0-9]*)', line)
                if match:
                    auto_what = match.group(1)
                    auto_name = match.group(2)
                    auto_where = f"{f.name}:{i + 1}"
                    auto_options = []
                    mode = "autodoc"
            elif mode == "autodoc":
                # blank or unindented line, back to text mode
                if not re.match(r'^\s\s', line):
                    yield auto_what, auto_name, auto_where, auto_options
                    auto_what = None
                    mode = "text"
                else:
                    # try to extract an option of the form '  :<name>:'
                    match = re.match(r"\s+:(\w+):", line)
                    if match:
                        auto_options.append(match.group(1))
            else:
                raise RuntimeError(f"unknown mode: {mode}")
        if auto_what:
            yield auto_what, auto_name, auto_where, auto_options


@pytest.mark.skipif(not has_docs(), reason="Docs not present")
def test_autodoc_has_noindex():
    bad = []  # is anyone truly bad or good? when it comes to autodocs: yes
    for p in docs_dir.glob("**/*.rst"):
        try:
            # If it is NOT a subdir of apidoc, raises ValueError
            p.relative_to(apidoc)
            # ignore anything under apidoc
        except ValueError:
            # not in apidoc, so look for autodoc sections
            for what, name, where, options in get_autodoc(p):
                if "noindex" not in options:
                    n = len(bad) + 1
                    bad.append(f"{n}) {what} '{name}' in {where}")
    if len(bad) > 0:
        bad_ones = "\n".join(bad)
        err = f"{len(bad)} 'autodoc' directive(s) missing 'noindex':\n{bad_ones}"
        assert False, err



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Project setup with setuptools
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_namespace_packages
import pathlib

cwd = pathlib.Path(__file__).parent.resolve()  # this will come in handy, probably

long_description = """WaterTAP is an open-source, integrated suite of predictive multi-scale models
for design and optimization of water treatment processes and systems. Specifically, WaterTAP is a new
library of water treatment-specific property, process unit, and network models that depend on the IDAES Platform,
an open source, next generation process systems engineering platform developed at the National Energy Technology
Laboratory with other partners. The WaterTAP project is funded by the NAWI  as a part of U.S. Department of
Energy’s Energy-Water Desalination Hub. The goal of WaterTAP is to assist the hub and the broader water R&D
community in assessing existing and emerging water treatment technologies by 1) providing predictive capabilities
involving the design, optimization, and performance of water treatment systems that will lead to improved energy
efficiency and lower cost, 2) advancing the state of the art for the design of water treatment components, systems
and networks to be comparable with, or even surpass, that in the chemical industry, and 3) disseminating these tools
for active use by water treatment researchers and engineers.""".replace(
    "\n", " "
).strip()


SPECIAL_DEPENDENCIES_FOR_RELEASE = [
    "idaes-pse==1.12.1",  # from PyPI
]

SPECIAL_DEPENDENCIES_FOR_PRERELEASE = [
    # update with a tag from the nawi-hub/idaes-pse
    # when a version of IDAES newer than the latest stable release from PyPI
    # will become needed for the watertap development
    "idaes-pse[prerelease] @ https://github.com/watertap-org/idaes-pse/archive/1.12.0.watertap.2022.01.06.zip",
]

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="watertap",
    url="https://github.com/watertap-org/watertap",
    version="0.3.1",
    description="WaterTAP modeling library",
    long_description=long_description,
    long_description_content_type="text/plain",
    author="NAWI team",
    license="BSD",
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="water systems, chemical engineering, process modeling, filtration, desalination, nawi",
    packages=find_namespace_packages(),
    python_requires=">=3.6, <4",
    install_requires=[
        # primary requirements for unit and property models
        # maintainers: switch to SPECIAL_DEPENDENCIES_FOR_RELEASE when cutting a release of watertap
        *SPECIAL_DEPENDENCIES_FOR_RELEASE,
        "pyomo>=6.2",  # (also needed for units in electrolyte database (edb))
        # the following requirements are for the electrolyte database (edb)
        "pymongo>3",  # database interface
        "fastjsonschema",  # schema validation
        "click",  # command-line tools with Click
        # tutorial tests
        "nbformat",
        "scipy",
        # https://www.python.org/dev/peps/pep-0508/#environment-markers
        'pywin32==225 ; platform_system=="Windows" and python_version>="3.8"',
    ],
    extras_require={
        "testing": [
            "pytest",
            "json-schema-for-humans",
            "mongomock",
        ],
        "dev": [
            "myst-parser",  # markdown support for Sphinx
            "nbsphinx",  # jupyter notebook support for sphinx
            "Sphinx",  # docs
            "sphinx_rtd_theme",  # docs
            # other requirements
            "linkify-it-py",
            "json-schema-for-humans",  # pretty JSON schema in HTML
            "black",  # code formatting
            # other requirements
            "pytest",  # test framework
            "pytest-cov",  # code coverage
            "mongomock", # mongodb mocking for testing
        ],
    },
    package_data={  # Optional
        "": [
            "*.json",
        ],
    },
    entry_points={
        # add edb CLI commands
        "console_scripts": [
            "edb = watertap.edb.commands:command_base",
        ]
    },
)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for validate module
"""
import pytest
from ..validate import validate
from .data import component_data, reaction_data


@pytest.mark.unit
@pytest.mark.parametrize("comp", component_data)
def test_validate_component(comp):
    validate(comp, obj_type="component")


@pytest.mark.unit
@pytest.mark.parametrize("reaction", reaction_data)
def test_validate_reaction(reaction):
    validate(reaction, obj_type="reaction")



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
eNRTL property configuration dicts for synthetic hard water

References:

[1] Islam, R.I., et al., Molecular thermodynamics for scaling prediction: Case
of membrane distillation, Separation and Purification Technology, 2021,
Vol. 276.

Author: Andrew Lee
"""

from pyomo.environ import Param, units as pyunits

from idaes.core import (AqueousPhase,
                        Solvent,
                        Apparent,
                        Anion,
                        Cation)
from idaes.generic_models.properties.core.eos.enrtl import ENRTL
from idaes.generic_models.properties.core.eos.enrtl_reference_states import \
    Symmetric, Unsymmetric
from idaes.generic_models.properties.core.generic.generic_property import (
        StateIndex)
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.pure.electrolyte import \
    relative_permittivity_constant


class ConstantVolMol():
    def build_parameters(b):
        b.vol_mol_pure = Param(initialize=18e-6,
                               units=pyunits.m**3/pyunits.mol,
                               mutable=True)

    def return_expression(b, cobj, T):
        return cobj.vol_mol_pure


configuration = {
    "components": {
        "H2O": {"type": Solvent,
                "vol_mol_liq_comp": ConstantVolMol,
                "relative_permittivity_liq_comp":
                    relative_permittivity_constant,
                "parameter_data": {
                    "mw": (18E-3, pyunits.kg/pyunits.mol),
                    "relative_permittivity_liq_comp": 78.54}},
        "NaCl": {"type": Apparent,
                 "dissociation_species": {"Na_+": 1, "Cl_-": 1}},
        "Na2SO4": {"type": Apparent,
                   "dissociation_species": {"Na_+": 2, "SO4_2-": 1}},
        "CaCl2": {"type": Apparent,
                  "dissociation_species": {"Ca_2+": 1, "Cl_-": 2}},
        "CaSO4": {"type": Apparent,
                  "dissociation_species": {"Ca_2+": 1, "SO4_2-": 1}},
        "MgCl2": {"type": Apparent,
                  "dissociation_species": {"Mg_2+": 1, "Cl_-": 2}},
        "MgSO4": {"type": Apparent,
                  "dissociation_species": {"Mg_2+": 1, "SO4_2-": 1}},
        "Na_+": {"type": Cation,
                 "charge": +1},
        "Ca_2+": {"type": Cation,
                  "charge": +2},
        "Mg_2+": {"type": Cation,
                  "charge": +2},
        "Cl_-": {"type": Anion,
                 "charge": -1},
        "SO4_2-": {"type": Anion,
                   "charge": -2}},
    "phases": {
        "Liq": {"type": AqueousPhase,
                "equation_of_state": ENRTL,
                "equation_of_state_options": {
                    "reference_state": Symmetric}}},
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "state_definition": FTPx,
    "state_components": StateIndex.true,
    "pressure_ref": 1e5,
    "temperature_ref": 298.15,
    "parameter_data": {
        "Liq_tau": {  # Table 1 [1]
            ('H2O', 'Na_+, Cl_-'): 8.866,
            ('Na_+, Cl_-', 'H2O'): -4.451,
            ('H2O', 'Ca_2+, Cl_-'): 10.478,
            ('Ca_2+, Cl_-', 'H2O'): -5.231,
            ('H2O', 'Mg_2+, Cl_-'): 10.854,
            ('Mg_2+, Cl_-', 'H2O'): -5.409,
            ('H2O', 'Na_+, SO4_2-'): 8.012,
            ('Na_+, SO4_2-', 'H2O'): -3.903,
            ('H2O', 'Ca_2+, SO4_2-'): 6.932,
            ('Ca_2+, SO4_2-', 'H2O'): -3.466,
            ('H2O', 'Mg_2+, SO4_2-'): 8.808,
            ('Mg_2+, SO4_2-', 'H2O'): -4.383,
            ('Na_+, Cl_-', 'Ca_2+, Cl_-'): -0.468,
            ('Ca_2+, Cl_-', 'Na_+, Cl_-'): 0.41,
            ('Na_+, Cl_-', 'Mg_2+, Cl_-'): -0.328,
            ('Mg_2+, Cl_-', 'Na_+, Cl_-'): -0.981,
            ('Ca_2+, Cl_-', 'Mg_2+, Cl_-'): 0.22,
            ('Mg_2+, Cl_-', 'Ca_2+, Cl_-'): 0.322,
            ('Na_+, SO4_2-', 'Ca_2+, SO4_2-'): -0.761,
            ('Ca_2+, SO4_2-', 'Na_+, SO4_2-'): 0.368,
            ('Na_+, SO4_2-', 'Mg_2+, SO4_2-'): -0.327,
            ('Mg_2+, SO4_2-', 'Na_+, SO4_2-'): 0.799,
            ('Ca_2+, SO4_2-', 'Mg_2+, SO4_2-'): 0,
            ('Mg_2+, SO4_2-', 'Ca_2+, SO4_2-'): 0.383,
            ('Na_+, Cl_-', 'Na_+, SO4_2-'): 0.803,
            ('Na_+, SO4_2-', 'Na_+, Cl_-'): -0.634,
            ('Ca_2+, Cl_-', 'Ca_2+, SO4_2-'): 0,
            ('Ca_2+, SO4_2-', 'Ca_2+, Cl_-'): -0.264,
            ('Mg_2+, Cl_-', 'Mg_2+, SO4_2-'): -0.707,
            ('Mg_2+, SO4_2-', 'Mg_2+, Cl_-'): -0.841
            }},
    "default_scaling_factors": {("mole_frac_comp", "Na_+"): 1e2,
                                ("mole_frac_comp", "Ca_2+"): 1e4,
                                ("mole_frac_comp", "Mg_2+"): 1e3,
                                ("mole_frac_comp", "SO4_2-"): 1e4,
                                ("mole_frac_comp", "Cl_-"): 1e2,
                                ("mole_frac_comp", "H2O"): 1,
                                ("flow_mol", None): 1e-1,
                                }
}



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for data_model module
"""
import logging
from pprint import pprint  # for debugging
import pytest
from . import data as testdata
from typing import Dict
from pyomo.environ import units as pyunits

# Used to test IDAES config -> DataWrapper
from idaes.generic_models.properties.core.pure import Perrys
from idaes.generic_models.properties.core.pure.NIST import NIST
from idaes.generic_models.properties.core.phase_equil.forms import fugacity
from idaes.generic_models.properties.core.reactions.equilibrium_forms import (
    log_power_law_equil,
)
from idaes.core import Component as IComponent
from idaes.generic_models.properties.core.reactions.equilibrium_constant import (
    van_t_hoff,
)
from idaes.generic_models.properties.core.reactions.dh_rxn import constant_dh_rxn
from idaes.generic_models.properties.core.generic.generic_reaction import (
    ConcentrationForm,
)
from idaes.core.components import Solvent, Solute, Cation, Anion


from ..data_model import (
    ConfigGenerator,
    Component,
    Reaction,
    Result,
    Base,
    ThermoConfig,
)

# For validating DataWrapper contents
from ..validate import validate

# errors
from ..error import BadConfiguration, ValidationError


@pytest.fixture
def debug_logging():
    """Use fixture cleanup ability to set and unset debug logging."""
    logging.getLogger("watertap.edb.data_model").setLevel(logging.DEBUG)
    yield "debug"
    logging.getLogger("watertap.edb.data_model").setLevel(logging.INFO)


def assert_configuration_equal(a: Dict, b: Dict, fld: str):
    """Walk through and compare things in two config dicts.

    This is needed because simple comparison will also compare objects that may not have appropriate
    equality methods defined and therefore fail for the simple reason that they have different object ids.
    """
    a_section, b_section = a.get(fld, {}), b.get(fld, {})
    assert len(a_section) == len(b_section)

    for name in a_section:
        assert name in b_section
        a_data, b_data = a_section[name], b_section[name]
        for key in a_data:
            assert key in b_data
            if key == "parameter_data":
                for key2, value2 in a_data[key].items():
                    assert key2 in b_data[key]
                    if (
                        isinstance(value2, tuple) and len(value2) == 2
                    ):  # number, unit pair
                        assert b_data[key][key2][0] == pytest.approx(value2[0])


@pytest.mark.unit
def test_config_generator():
    hc = ConfigGenerator({})


@pytest.mark.unit
def test_component_ca_thermo():
    logging.getLogger("idaes.watertap.edb.data_model").setLevel(logging.DEBUG)
    comp = Component(testdata.Ca_thermo_data)
    generated_config = comp.idaes_config

    # for debugging
    print("Config generated from data:")
    pprint(generated_config)
    print("Expected idaes_config:")
    pprint(testdata.Ca_thermo_config)

    assert_configuration_equal(
        comp.idaes_config, testdata.Ca_thermo_config, "components"
    )
    logging.getLogger("idaes.watertap.edb.data_model").setLevel(logging.INFO)


@pytest.mark.unit
def test_reaction_bicarbonate():
    print(testdata.bicarbonate_reaction_data)
    react = Reaction(testdata.bicarbonate_reaction_data)
    generated_config = react.idaes_config

    # for debugging
    print("Config generated from data:")
    pprint(generated_config)
    print("Expected idaes_config:")
    pprint(testdata.bicarbonate_reaction_config)

    assert_configuration_equal(
        generated_config, testdata.bicarbonate_reaction_config, "equilibrium_reactions"
    )


@pytest.mark.unit
@pytest.mark.parametrize("data", testdata.reaction_data)
def test_reaction(data):
    reaction = Reaction(data)


@pytest.mark.unit
def test_result():
    Result(iterator=[], item_class=Component)
    Result(iterator=[], item_class=Reaction)


@pytest.mark.unit
@pytest.mark.parametrize(
    "data_wrapper_class,required",
    [
        (
            Component,
            {"name": "foo", "elements": ["H"], "type": "solvent", "parameter_data": {}},
        ),
        (
            Reaction,
            {
                "name": "foo",
                "type": "equilibrium",
                "parameter_data": {},
                "components": ["H2O"],
            },
        ),
    ],
)
def test_config_generator_required(data_wrapper_class, required):
    # make sure it fails with any missing required field(s)
    all_fields = list(required.keys())
    for i in range(len(all_fields)):
        except_i = all_fields[:i] + all_fields[i + 1 :]
        missing_a_field = {f: required[f] for f in except_i}
        print(f"Data missing field '{all_fields[i]}': {missing_a_field}")
        if data_wrapper_class is Component and all_fields[i] in ("elements", "type"):
            # 'elements' field will be filled by pre-processor for components
            _ = data_wrapper_class(missing_a_field).idaes_config
        else:
            with pytest.raises(ValidationError):
                _ = data_wrapper_class(missing_a_field).idaes_config
    # add all required fields and make sure it passes
    d = {r: required[r] for r in required}
    _ = data_wrapper_class(d).idaes_config


@pytest.mark.unit
@pytest.mark.parametrize("starting_value", [None, {}])
def test_base(starting_value):
    starting = dict.fromkeys(Component.merge_keys, starting_value)
    foo_value = 12.34
    mk0 = Component.merge_keys[0]
    starting[mk0] = {"foo": foo_value}
    b = Base(starting)
    assert b.idaes_config == starting
    # Add an empty component
    with pytest.raises(BadConfiguration):
        c = Component({})  # "name" is required
    # also required: elements, parameter_data
    c = Component({"name": "a", "elements": ["H +"], "parameter_data": {}})
    b.add(c)
    assert b.idaes_config[mk0]["foo"] == starting[mk0]["foo"]
    # Add a non-empty component
    name = "baz"
    component_data = {
        "name": name,
        "elements": [],
        "parameter_data": {"foo_coeff": [{"i": 0, "v": 1, "u": "g"}]},
    }
    c = Component(component_data)
    b.add(c)
    print(f"b.idaes_config={b.idaes_config} component_data={component_data}")
    config_foo_coeff_value = b.idaes_config[mk0][name]["parameter_data"]["foo_coeff"][0]
    data_foo_coeff_value = component_data["parameter_data"]["foo_coeff"][0]["v"]
    assert config_foo_coeff_value == data_foo_coeff_value


subst_foo, subst_bar, subst_y1, subst_y2 = "foo_obj", "bar_obj", 1, 2


class SubstituteTestGenerator(ConfigGenerator):
    pass


@pytest.mark.unit
def test_config_generator_substitute(debug_logging):
    SubstituteTestGenerator.substitute_values = {
        "a.b": {"foo": subst_foo},
        "x.*_bla": {"foo": subst_foo, "bar": subst_bar},
        "y": {"1": subst_y1, "2": subst_y2},
        "d.e.e.p": {"number_one": subst_y1},
        "d.e.e.p.e.*": {"number_two": subst_y2},
        "z.*": SubstituteTestGenerator.SUBST_UNITS,
    }
    data = {
        "a": {"b": "foo"},
        "x": {"one_bla": "bar", "two_bla": "hello", "three_bla": "foo", "ignore": "me"},
        "y": "1",
        "z": {"time": "s", "ignored_due_to_value": 0.123},
        "d": {"e": {"e": {"p": "number_one"}}},
    }
    print(f"before: {data}")
    SubstituteTestGenerator._substitute(data)
    print(f"after: {data}")
    assert data == {
        "a": {"b": "foo_obj"},
        "x": {
            "one_bla": "bar_obj",
            "two_bla": "hello",
            "three_bla": "foo_obj",
            "ignore": "me",
        },
        "y": 1,
        "z": {"time": pyunits.s, "ignored_due_to_value": 0.123},
        "d": {"e": {"e": {"p": 1}}},
    }


def test_substitute_phases(debug_logging):
    data = {"phases": {"Liq": {"type": "LiquidPhase"}}}
    SubstituteTestGenerator.substitute_values = {
        "phases.Liq.type": {"LiquidPhase": NIST}
    }
    print(f"before: {data}")
    SubstituteTestGenerator._substitute(data)
    print(f"after: {data}")
    assert data["phases"]["Liq"]["type"] == NIST


def test_component_from_idaes_config(debug_logging):
    H2O_thermo_config = {
        "components": {
            "H2O": {
                "type": IComponent,
                # Define the methods used to calculate the following properties
                "dens_mol_liq_comp": Perrys,
                "enth_mol_liq_comp": Perrys,
                "cp_mol_liq_comp": Perrys,
                "entr_mol_liq_comp": Perrys,
                "enth_mol_ig_comp": NIST,
                "pressure_sat_comp": NIST,
                "phase_equilibrium_form": {("Vap", "Liq"): fugacity},
                # Parameter data is always associated with the methods defined above
                "parameter_data": {
                    "mw": (18.0153, pyunits.g / pyunits.mol),
                    "pressure_crit": (220.64e5, pyunits.Pa),
                    "temperature_crit": (647, pyunits.K),
                    # Comes from Perry's Handbook:  p. 2-98
                    "dens_mol_liq_comp_coeff": {
                        "1": (5.459, pyunits.kmol * pyunits.m ** -3),
                        "2": (0.30542, pyunits.dimensionless),
                        "3": (647.13, pyunits.K),
                        "4": (0.081, pyunits.dimensionless),
                    },
                    "enth_mol_form_liq_comp_ref": (-285.830, pyunits.kJ / pyunits.mol),
                    "enth_mol_form_vap_comp_ref": (0, pyunits.kJ / pyunits.mol),
                    # Comes from Perry's Handbook:  p. 2-174
                    "cp_mol_liq_comp_coeff": {
                        "1": (2.7637e5, pyunits.J / pyunits.kmol / pyunits.K),
                        "2": (-2.0901e3, pyunits.J / pyunits.kmol / pyunits.K ** 2),
                        "3": (8.125, pyunits.J / pyunits.kmol / pyunits.K ** 3),
                        "4": (-1.4116e-2, pyunits.J / pyunits.kmol / pyunits.K ** 4),
                        "5": (9.3701e-6, pyunits.J / pyunits.kmol / pyunits.K ** 5),
                    },
                    "cp_mol_ig_comp_coeff": {
                        "A": (30.09200, pyunits.J / pyunits.mol / pyunits.K),
                        "B": (
                            6.832514,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** -1,
                        ),
                        "C": (
                            6.793435,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** -2,
                        ),
                        "D": (
                            -2.534480,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** -3,
                        ),
                        "E": (
                            0.082139,
                            pyunits.J
                            * pyunits.mol ** -1
                            * pyunits.K ** -1
                            * pyunits.kiloK ** 2,
                        ),
                        "F": (-250.8810, pyunits.kJ / pyunits.mol),
                        "G": (223.3967, pyunits.J / pyunits.mol / pyunits.K),
                        "H": (0, pyunits.kJ / pyunits.mol),
                    },
                    "entr_mol_form_liq_comp_ref": (
                        69.95,
                        pyunits.J / pyunits.K / pyunits.mol,
                    ),
                    "pressure_sat_comp_coeff": {
                        "A": (4.6543, None),  # [1], temperature range 255.9 K - 373 K
                        "B": (1435.264, pyunits.K),
                        "C": (-64.848, pyunits.K),
                    },
                },
            }
        }
    }
    logging.getLogger("watertap.edb.data_model").setLevel(logging.DEBUG)
    # create Component from the config
    components = Component.from_idaes_config(H2O_thermo_config)
    assert len(components) == 1
    component = components[0]
    # make sure it is a valid Component
    validate(component)
    # create a config from the Component, i.e. round-trip, and compare
    assert_configuration_equal(H2O_thermo_config, component.idaes_config, "components")
    logging.getLogger("watertap.edb.data_model").setLevel(logging.INFO)


def test_reaction_from_idaes_config(debug_logging):
    carbonation_reaction_config = {
        "base_units": {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        },
        "equilibrium_reactions": {
            "CO2_to_H2CO3": {
                "stoichiometry": {
                    ("Liq", "H2O"): -1,
                    ("Liq", "CO2"): -1,
                    ("Liq", "H2CO3"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                # "equilibrium_constant": gibbs_energy,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.molarity,
                "parameter_data": {
                    "dh_rxn_ref": (0, pyunits.kJ / pyunits.mol),
                    # NOTE: This 'ds_rxn_ref' was calculated to give 'keq' of 1.7*10**-3
                    # "ds_rxn_ref": (-53.0, pyunits.J/pyunits.mol/pyunits.K),
                    "k_eq_ref": (1.7 * 10 ** -3, None),
                    "T_eq_ref": (300, pyunits.K),
                    "reaction_order": {
                        ("Liq", "H2CO3"): 1,
                        ("Liq", "CO2"): -1,
                        ("Liq", "H2O"): 0,
                    },
                },
            }
        },
    }
    # create Component from the config
    result = Reaction.from_idaes_config(carbonation_reaction_config)
    assert len(result) == 1
    reaction = result[0]
    # make sure it is a valid Reaction
    validate(reaction)

    # in the generated config 'reaction_order' will be dropped since it is a
    # runtime addition, not something we store in the DB. So for comparison
    # purposes drop it from the original config as well
    del carbonation_reaction_config["equilibrium_reactions"]["CO2_to_H2CO3"][
        "parameter_data"
    ]["reaction_order"]

    # create a config from the Reaction, i.e. round-trip, and compare
    assert_configuration_equal(
        carbonation_reaction_config, reaction.idaes_config, "equilibrium_reactions"
    )


def test_thermoconfig_set_type():
    def _type(d):
        keys = list(d["components"].keys())
        return d["components"][keys[0]]["type"]

    # water -> solvent
    data = {
        "parameter_data": {
            "mw": [{"v": 18.0153, "u": "g/mol", "i": 0}],
        },
        "name": "H2O",
        "elements": ["O", "H"],
    }
    thermo = Component(data, validation=False)
    config = thermo.idaes_config
    assert _type(config) == Solvent

    # neutral non-water -> solute
    data = {
        "parameter_data": {
            "mw": [{"v": 96.994, "u": "g/mol", "i": 0}],
        },
        "name": "H2PO4",
        "elements": ["O", "H", "P"],
    }
    thermo = Component(data)
    config = thermo.idaes_config
    assert _type(config) == Solute

    # positive charge -> cation
    data = {
        "parameter_data": {
            "mw": [{"v": 57.002, "u": "g/mol", "i": 0}],
        },
        "name": "CaOH +",
        "elements": ["Ca", "O", "H"],
    }
    thermo = Component(data)
    config = thermo.idaes_config
    assert _type(config) == Cation

    # negative charge -> anion
    data = {
        "parameter_data": {
            "mw": [{"v": 17.008, "u": "g/mol", "i": 0}],
        },
        "name": "OH -",
        "elements": ["O", "H"],
    }
    thermo = Component(data)
    config = thermo.idaes_config
    assert _type(config) == Anion

    # already present, even if wrong, leave alone
    data = {
        "type": "solvent",
        "parameter_data": {
            "mw": [{"v": 17.008, "u": "g/mol", "i": 0}],
        },
        "name": "OH -",
        "elements": ["O", "H"],
    }
    thermo = Component(data)
    config = thermo.idaes_config
    assert _type(config) == Solvent


@pytest.mark.unit
def test_reaction_order():
    # minimal Reaction object to work with
    def reaction():
        return Reaction(
        {
            "name": "foo",
            "components": [],
            "elements": ["Ca", "O", "H"],
            # valid chemistry? no. useful? yes.
            Reaction.NAMES.param: {
                "reaction_order": {
                    "Liq": {"B": 2, "C": 1, "H": 1},
                    "Vap": {"B": 1, "C": -2, "H": 1},
                    "Sol": {"B": -1, "C": 2, "H": 0},
                }
            },
            "type": "equilibrium",
        }
        )

    # the error you get when the input doesn't pass JSON schema validation
    bad_input_error = ValidationError
    # Reaction missing param
    with pytest.raises(bad_input_error):
        Reaction({"elements": ["H"]}).set_reaction_order("Foo", {})
    # Reaction missing reaction_order, still OK handled in pre-processing
    # but empty input list causes an error
    with pytest.raises(ValueError):
        Reaction({"components": [], "elements": ["H"], Reaction.NAMES.param: {},
                  "type": "equilibrium", "name": "Foo"}).set_reaction_order(
                    "Foo", {})
    # Invalid phase
    r = reaction()
    with pytest.raises(ValueError):  # bad phase
        r.set_reaction_order("Plasma", {"B": 1, "C": 1, "H": 1})
    # No components
    with pytest.raises(ValueError):
        r.set_reaction_order("Liq", {})
    with pytest.raises(ValueError):
        r.set_reaction_order("Liq", [])
    # Bad components
    with pytest.raises(KeyError):
        r.set_reaction_order("Liq", {"X": 3})
    with pytest.raises(KeyError):
        r.set_reaction_order("Liq", [("X", 3)])
    with pytest.raises(ValueError):
        r.set_reaction_order("Liq", {"B": 1}, require_all=True)
    with pytest.raises(ValueError):
        r.set_reaction_order("Liq", [("B", 0), ("C", 0), ("X", 0)], require_all=True)
    # check happy path
    values = [{"B": 0.5, "C": 1.5}, [("B", -2), ("C", 2)]]
    for i in range(2):
        r = reaction()
        v = values[i]
        r.set_reaction_order("Liq", v)
        ro = r.data[Reaction.NAMES.param][Reaction.NAMES.reaction_order]["Liq"]
        for k in v:
            if i == 0:
                assert ro[k] == v[k]
            else:
                assert ro[k[0]] == dict(v)[k[0]]
        r.set_reaction_order("Sol", v)
        ro = r.data[Reaction.NAMES.param][Reaction.NAMES.reaction_order]["Sol"]
        for k in v:
            if i == 0:
                assert ro[k] == v[k]
            else:
                assert ro[k[0]] == dict(v)[k[0]]



import logging
from pathlib import Path
from watertap.edb import ElectrolyteDB
import pytest

_log = logging.getLogger(__name__)

DOCS_DIR = "docs"  # expected documentation directory name


def get_docs_root():
    start = Path(__file__)
    p, pp = start, start.parent
    # loop while there is a parent and it's not named 'site_packages'
    while pp != p and pp.name != "site_packages":
        target = p / DOCS_DIR
        if target.exists() and target.is_dir():
            return start, target
        p, pp = pp, pp.parent
    # not found
    return start, None


@pytest.fixture(scope="session")
def docs_root():
    """Find docs root, or call pytest.skip
    """
    start, result = get_docs_root()
    if result is None:
        pytest.skip(f"No directory '{DOCS_DIR}' found from '{start}'")
    yield result


def check_for_mongodb() -> bool:
    try:
        edb = ElectrolyteDB()  # will fail if no DB
        edb.get_base()  # will fail if DB is not loaded
    except Exception as err:
        _log.warning(f"Could not connect to MongoDB: {err}")
    return False


@pytest.fixture(scope="module")
def electrolytedb():
    """See if EDB can be instantiated, or call pytest.skip
    """
    if not check_for_mongodb():
        pytest.skip("MongoDB is required")



import sys
import os
import time

from watertap.tools.parameter_sweep import _init_mpi, LinearSample, parameter_sweep


def append_costing_outputs(m, outputs, units_to_cost):
    for unit_name in units_to_cost:
        for cost_type in ['capital_cost', 'operating_cost']:
            unit = getattr(m.fs, unit_name)
            cost = getattr(unit.costing, cost_type)

            outputs['%s_%s' % (cost_type, unit_name)] = cost

    return outputs


def run_analysis(case_num, nx, RO_type):

    desal_kwargs = {'has_desal_feed': False, 'is_twostage': True, 'has_ERD': True,
                    'RO_type': RO_type, 'RO_base': 'TDS', 'RO_level': 'detailed'}
    
    sweep_params = {}
    outputs = {}
    optimize_kwargs=None

    if case_num == 1:
        # ================================================================
        # Single Stage RO
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_single_stage as fs_single_stage

        desal_kwargs['is_twostage'] = False

        m = fs_single_stage.optimize_flowsheet(**desal_kwargs)

        sweep_params['System Recovery'] = LinearSample(m.fs.system_recovery_target, 0.3, 0.95, nx)
        # sweep_params['RO Recovery'] = LinearSample(m.fs.RO_recovery, 0.3, 0.95, nx)

        outputs['LCOW'] = m.fs.costing.LCOW
        outputs['Saturation Index'] = m.fs.desal_saturation.saturation_index
        outputs['Pump Pressure'] = m.fs.pump_RO.control_volume.properties_out[0].pressure
        outputs['RO Recovery'] = m.fs.RO_recovery
        outputs['Annual Water Production'] = m.fs.annual_water_production
        outputs = append_costing_outputs(m, outputs, ['RO', 'pump_RO', 'ERD'])

        output_filename = 'output/fs_single_stage/results_%d_%sRO.csv' % (case_num, desal_kwargs['RO_type'])

        opt_function = fs_single_stage.optimize

    elif case_num == 2:
        # ================================================================
        # Two Stage RO
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_two_stage as fs_two_stage

        m = fs_two_stage.optimize_flowsheet(**desal_kwargs)

        sweep_params['System Recovery'] = LinearSample(m.fs.system_recovery_target, 0.3, 0.95, nx)
        # sweep_params['RO Recovery'] = LinearSample(m.fs.RO_recovery, 0.3, 0.95, nx)
        sweep_params['Max Pressure'] = LinearSample(m.fs.max_allowable_pressure, 300e5, 75e5, nx)

        outputs['LCOW'] = m.fs.costing.LCOW
        outputs['Saturation Index'] = m.fs.desal_saturation.saturation_index
        outputs['RO-1 Pump Pressure'] = m.fs.pump_RO.control_volume.properties_out[0].pressure
        outputs['RO-2 Pump Pressure'] = m.fs.pump_RO2.control_volume.properties_out[0].pressure
        outputs['RO Recovery'] = m.fs.RO_recovery
        outputs['Annual Water Production'] = m.fs.annual_water_production
        outputs = append_costing_outputs(m, outputs, ['RO', 'pump_RO', 'RO2', 'pump_RO2', 'ERD'])

        output_filename = 'output/fs_two_stage/results_%d_%sRO_no_sat.csv' % (case_num, desal_kwargs['RO_type'])

        opt_function = fs_two_stage.optimize

    elif case_num == 3:
        # ================================================================
        # NF No Bypass
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_NF_no_bypass as fs_NF_no_bypass

        m = fs_NF_no_bypass.solve_flowsheet()

        sweep_params['NF Recovery'] = LinearSample(m.fs.NF.recovery_mass_phase_comp[0, 'Liq', 'H2O'], .3, .95, nx)

        outputs['LCOW'] = m.fs.costing.LCOW
        outputs['Saturation Index'] = m.fs.pretrt_saturation.saturation_index

        output_filename = 'output/fs_NF_no_bypass/results_%d.csv' % case_num

        opt_function = fs_NF_no_bypass.simulate
        # Need to unfix NF area to simulate with fixed NF recovery
        optimize_kwargs = {'unfix_nf_area': True}

    elif case_num == 4:
        # ================================================================
        # NF Two Stage
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_NF as fs_NF

        m = fs_NF.optimize_flowsheet()

        sweep_params['NF Split Fraction'] = LinearSample(m.fs.splitter.split_fraction[0, 'pretreatment'], .01, .99, nx)

        outputs['LCOW']  = m.fs.costing.LCOW
        outputs['Ca Removal'] = m.fs.removal_Ca
        outputs['Max Concentration Constraint']  = m.fs.eq_max_conc_NF.body

        output_filename = 'output/fs_NF/results_%d_%sRO.csv' % (case_num, desal_kwargs['RO_type'])

        opt_function = fs_NF.optimize

    elif case_num == 5:
        # ================================================================
        # NF Two Stage
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_NF_two_stage as fs_NF_two_stage

        m = fs_NF_two_stage.optimize_flowsheet(**desal_kwargs)

        sweep_params['NF Split Fraction'] = LinearSample(m.fs.splitter.split_fraction[0, 'pretreatment'], 0.25, 0.99, 4)
        # sweep_params['RO Recovery'] = LinearSample(m.fs.RO_recovery, 0.3, 0.95, nx)
        sweep_params['System Recovery'] = LinearSample(m.fs.system_recovery_target, 0.5, 0.95, nx)

        outputs['LCOW'] = m.fs.costing.LCOW
        outputs['RO Recovery'] = m.fs.RO_recovery

        output_filename = 'output/fs_NF_two_stage/results_%d_%sRO.csv' % (case_num, desal_kwargs['RO_type'])

        opt_function = fs_NF_two_stage.optimize

    elif case_num == 6:
        # ================================================================
        # NF Two Stage
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_NF_two_stage as fs_NF_two_stage

        m = fs_NF_two_stage.optimize_flowsheet(**desal_kwargs)

        sweep_params['System Recovery'] = LinearSample(m.fs.system_recovery_target, 0.5, 0.95, nx)

        outputs['LCOW'] = m.fs.costing.LCOW
        outputs['NF Split Fraction'] = m.fs.splitter.split_fraction[0, 'pretreatment']

        output_filename = 'output/fs_NF_two_stage/results_%d_%sRO.csv' % (case_num, desal_kwargs['RO_type'])

        opt_function = fs_NF_two_stage.optimize

    elif case_num == 7:
        # ================================================================
        # NF Two Stage
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_NF_two_stage as fs_NF_two_stage

        m = fs_NF_two_stage.optimize_flowsheet(**desal_kwargs)

        sweep_params['NF Split Fraction'] = LinearSample(m.fs.splitter.split_fraction[0, 'pretreatment'], 0.25, 0.99, 4)
        sweep_params['System Recovery'] = LinearSample(m.fs.system_recovery_target, 0.5, 0.95, nx)

        outputs['LCOW'] = m.fs.costing.LCOW

        output_filename = 'output/fs_NF_two_stage/results_%d_%sRO.csv' % (case_num, desal_kwargs['RO_type'])

        opt_function = fs_NF_two_stage.optimize

    elif case_num == 8:
        # ================================================================
        # Softening
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_softening as fs_softening

        m = fs_softening.solve_flowsheet()

        sweep_params['Lime Dosing'] = LinearSample(m.fs.stoich_softening_mixer_unit.lime_stream_state[0].flow_mol, 0.0, 0.128, nx)

        outputs['LCOW'] = m.fs.costing.LCOW
        outputs['Ca Removal'] = m.fs.removal_Ca
        outputs['Mg Removal'] = m.fs.removal_Mg
        outputs['Annual Water Production'] = m.fs.annual_water_production
        outputs['capital_cost_total'] = m.fs.costing.capital_cost_total
        outputs['operating_cost_total'] = m.fs.costing.operating_cost_total

        output_filename = 'output/fs_softening/results_%d.csv' % case_num

        opt_function = fs_softening.simulate

    elif case_num == 9:
        # ================================================================
        # Softening Two Stage
        # ================================================================
        import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_softening_two_stage as fs_softening_two_stage

        m = fs_softening_two_stage.optimize_flowsheet(**desal_kwargs)

        sweep_params['System Recovery'] = LinearSample(m.fs.system_recovery_target, 0.5, 0.95, nx)

        outputs['LCOW'] = m.fs.costing.LCOW
        outputs['Lime Dosing'] = m.fs.stoich_softening_mixer_unit.lime_stream_state[0].flow_mol

        output_filename = 'output/fs_softening_two_stage/results_%d_%sRO.csv' % (case_num, desal_kwargs['RO_type'])

        opt_function = fs_softening_two_stage.optimize

    else:
        raise ValueError('case_num = %d not recognized.' % (case_num))


    global_results = parameter_sweep(m, sweep_params, outputs, output_filename, 
                                     optimize_function=opt_function,
                                     optimize_kwargs=optimize_kwargs,
                                     debugging_data_dir=os.path.split(output_filename)[0]+'/local',
                                     interpolate_nan_outputs=True)

    return global_results, sweep_params


if __name__ == "__main__":

    # Start MPI communicator
    comm, rank, num_procs = _init_mpi()

    # Get the case number to run
    try:
        case_num = int(sys.argv[1])
    except:
        # Default to running case 1
        case_num = 1

    # Get the default number of discretization points
    try:
        nx = int(sys.argv[2])
    except:
        # Default to a 4-point discretization
        nx = 4

    # Get the default RO unit to use (1D or 0D)
    try:
        RO_type = sys.argv[3]
    except:
        # Default to RO_type='0D'
        RO_type = '0D'

    tic = time.time()
    global_results, sweep_params = run_analysis(case_num, nx, RO_type)
    print(global_results)
    toc = time.time()

    if rank == 0:
        total_samples = 1

        for k, v in sweep_params.items():
            total_samples *= v.num_samples
            
        print('Finished case_num = %d.' % (case_num))
        print('Processed %d swept parameters comprising %d total points.' % (len(sweep_params), total_samples))
        print('Elapsed time = %.1f s.' % (toc-tic))




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
'''
mutable parameters for optimization:
    m.fs.system_recovery_target
    m.fs.max_allowable_pressure
'''
from pyomo.environ import ConcreteModel, TransformationFactory, Constraint, Param

from idaes.core import FlowsheetBlock
from idaes.core.util.scaling import calculate_scaling_factors

from watertap.examples.flowsheets.full_treatment_train.util import (solve_with_user_scaling,
                                                             check_dof)

import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_NF as flowsheet_NF
import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_two_stage as flowsheet_two_stage


def build(m, **kwargs):
    """
    Build a flowsheet with softening pretreatment and RO.
    """

    assert kwargs['RO_base'] == 'TDS'
    assert not kwargs['has_desal_feed']
    flowsheet_NF.build_components(m)
    kwargs['NF_type'] = 'ZO'
    flowsheet_two_stage.build_components(m, pretrt_type='NF', **kwargs)

    m.fs.feed.properties[0].flow_vol
    m.fs.feed.properties[0].conc_mol_phase_comp['Liq', 'Ca']

    m.fs.tb_pretrt_to_desal.properties_in[0].flow_vol
    m.fs.tb_pretrt_to_desal.properties_in[0].conc_mol_phase_comp['Liq', 'Ca']

    return m


def scale(m, **kwargs):
    flowsheet_NF.scale(m)
    flowsheet_two_stage.scale(m, **kwargs)


def initialize(m, **kwargs):
    flowsheet_NF.initialize(m)
    flowsheet_two_stage.initialize(m, **kwargs)


def report(m, **kwargs):
    flowsheet_NF.report(m)
    flowsheet_two_stage.report(m, **kwargs)


def set_optimization_components(m, system_recovery, **kwargs):
    # unfix variables
    m.fs.splitter.split_fraction[0, 'bypass'].unfix()
    m.fs.splitter.split_fraction[0, 'bypass'].setlb(0.001)
    m.fs.splitter.split_fraction[0, 'bypass'].setub(0.99)

    m.fs.NF.area.unfix()
    m.fs.NF.area.setlb(0.1)
    m.fs.NF.area.setub(1000)

    m.fs.max_conc_factor_target = Param(initialize=3.5, mutable=True)
    m.fs.max_conc_factor_target_tol = Param(initialize=5e-6, mutable=True)
    m.fs.eq_max_conc_NF = Constraint(
        expr=(0,
              m.fs.NF.feed_side.properties_out[0].mass_frac_phase_comp['Liq', 'Ca']
              - m.fs.max_conc_factor_target * m.fs.feed.properties[0].mass_frac_phase_comp['Liq', 'Ca'],
              m.fs.max_conc_factor_target_tol))

    flowsheet_two_stage.set_optimization_components(m, system_recovery, **kwargs)


def set_up_optimization(m, system_recovery=0.50, **kwargs):
    set_optimization_components(m, system_recovery, **kwargs)
    calculate_scaling_factors(m)
    check_dof(m, 8)


def optimize(m):
    solve_with_user_scaling(m, tee=False, fail_flag=True)


def solve_flowsheet(**desal_kwargs):
    if desal_kwargs == {}:
        desal_kwargs = flowsheet_two_stage.desal_kwargs

    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    build(m, **desal_kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    scale(m, **desal_kwargs)
    calculate_scaling_factors(m)

    # initialize
    initialize(m, **desal_kwargs)

    check_dof(m)
    solve_with_user_scaling(m, tee=False, fail_flag=True)

    # report
    print('==================================='
          '\n          Simulation          ')
    report(m, **desal_kwargs)

    return m


def optimize_flowsheet(system_recovery=0.50, **kwargs):
    m = solve_flowsheet(**kwargs)
    set_up_optimization(m, system_recovery=system_recovery, **kwargs)
    optimize(m)

    print('==================================='
          '\n       Optimization            ')
    report(m, **flowsheet_two_stage.desal_kwargs)

    return m


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        m = solve_flowsheet()
    else:
        desal_kwargs = flowsheet_two_stage.desal_kwargs
        m = optimize_flowsheet(system_recovery=float(sys.argv[1]), **desal_kwargs)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
This is the EDB version of test_enrtl_water_pH.py

See that module for details on what the tests accomplish
in terms of the IDAES chemistry packages.

XXX: This test currently doesn't *do* anything
"""
import pytest
from watertap.edb import ElectrolyteDB

# Set global database object after checking that MongoDB server is up
g_edb = None
if ElectrolyteDB.can_connect():
    g_edb = ElectrolyteDB()


def get_thermo_config(edb):
    from idaes.generic_models.properties.core.eos.enrtl import ENRTL
    from idaes.generic_models.properties.core.eos.enrtl_reference_states import (
        Unsymmetric,
    )

    base = edb.get_base("default_thermo")
    elements = ["H", "O"]
    components = []
    # Add the components
    for c in edb.get_components(element_names=elements):
        # Need to remove these to avoid errors when using the generated config
        c.remove("valid_phase_types")
        c.remove("enth_mol_ig_comp")
        c.remove("phase_equilibrium_form")
        c.remove("pressure_sat_comp")
        # Add specific stuff not in the DB
        if c.name == "H2O":
            c.data["relative_permittivity_liq_comp"] = "relative_permittivity_constant"
            c.set_parameter("relative_permittivity_liq_comp", 78.54)
        base.add(c)
        components.append(c.name)

    # Add the reaction
    r = list(edb.get_reactions(reaction_names=["H2O_Kw_2"]))[0]
    r.set_reaction_order("Liq", ("H2O",), ("H_+", "OH_-"))
    base.add(r)

    cfg = base.idaes_config.copy()
    cfg["phases"]["Liq"]["equation_of_state"] = ENRTL
    cfg["phases"]["Liq"]["equation_of_state_options"] = {"reference_state": Unsymmetric}
    cfg["inherent_reactions"] = cfg["equilibrium_reactions"]
    del cfg["equilibrium_reactions"]
    return cfg


def get_water_reaction_config(edb):
    from idaes.generic_models.properties.core.reactions.equilibrium_forms import (
        log_power_law_equil,
    )

    elements = ["H", "O"]
    components = [c.name for c in edb.get_components(element_names=elements)]
    base = edb.get_base("default_thermo")
    # Just use base units, and add a dummy equilibrium reaction config
    cfg = {
        "base_units": base.idaes_config["base_units"],
        "equilibrium_reactions": {
            "dummy": {
                "stoichiometry": {},
                "equilibrium_form": log_power_law_equil,
            }
        },
    }
    return cfg


# ====================================================================================


def get_carbonic_thermo_config(edb):
    from idaes.generic_models.properties.core.eos.enrtl import ENRTL
    from idaes.generic_models.properties.core.eos.enrtl_reference_states import (
        Unsymmetric,
    )

    base = edb.get_base("default_thermo")
    # The 'right' way to fetch all components
    # elements = ["H", "O", "C", "Na"]
    components = []
    # Add the components
    component_names = ("Na_+", "CO3_2-",  "HCO3_-", "H2CO3", "OH_-", "H_+", "H2O")
    for c in edb.get_components(component_names=component_names):
        # Need to remove these to avoid errors when using the generated config
        if c.name != "H2CO3":
            c.remove("valid_phase_types")
        c.remove("enth_mol_ig_comp")
        c.remove("phase_equilibrium_form")
        c.remove("pressure_sat_comp")
        # Add specific stuff not in the DB
        if c.name == "H2O":
            c.data["relative_permittivity_liq_comp"] = "relative_permittivity_constant"
            c.set_parameter("relative_permittivity_liq_comp", 78.54)
        base.add(c)
        components.append(c.name)

    # Add the reaction
    for r in edb.get_reactions(reaction_names=["H2O_Kw_2", "H2CO3_Ka1", "H2CO3_Ka2"]):
        if r.name == "H2O_Kw_2":
            r.set_reaction_order("Liq", ("H2O",), ("H_+", "OH_-"))
        elif r.name == "H2CO3_Ka1":
            r.set_reaction_order("Liq", ("H2CO3",), ("H_+", "HCO3_-"), lhs_value=-1)
            r.remove_parameter("ds_rxn_ref")
            r.set_parameter("dh_rxn_ref", 0, units="kJ/mol")
        elif r.name == "H2CO3_Ka2":
            r.set_reaction_order("Liq", ("HCO3_-",), ("H_+", "CO3_2-"), lhs_value=-1)
            r.remove_parameter("ds_rxn_ref")
            r.set_parameter("dh_rxn_ref", 0, units="kJ/mol")
        base.add(r)

    cfg = base.idaes_config.copy()
    cfg["phases"]["Liq"]["equation_of_state"] = ENRTL
    cfg["phases"]["Liq"]["equation_of_state_options"] = {"reference_state": Unsymmetric}
    cfg["inherent_reactions"] = cfg["equilibrium_reactions"]
    del cfg["equilibrium_reactions"]
    return cfg



###############################################################################
# ProteusLib Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/nawi-hub/proteuslib/"
#
###############################################################################

import pytest
from pyomo.environ import ConcreteModel, assert_optimal_termination, value
from idaes.core import FlowsheetBlock
import idaes.core.util.scaling as iscale
from pyomo.util.check_units import assert_units_consistent
from idaes.generic_models.properties.core.generic.generic_property import GenericParameterBlock
from watertap.property_models.seawater_ion_generic import configuration
from watertap.core.util.initialization import check_dof
from idaes.core.util import get_solver

solver = get_solver()
# -----------------------------------------------------------------------------

@pytest.mark.component
def test_property_seawater_ions():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = GenericParameterBlock(default=configuration)
    m.fs.stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    # specify
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'Na_+'].fix(0.008845)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'Ca_2+'].fix(0.000174)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'Mg_2+'].fix(0.001049)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'SO4_2-'].fix(0.000407)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'Cl_-'].fix(0.010479)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'H2O'].fix(0.979046)
    m.fs.stream[0].temperature.fix(273.15 + 25)
    m.fs.stream[0].pressure.fix(101325)


    # # scaling
    iscale.calculate_scaling_factors(m.fs)

    # checking state block
    assert_units_consistent(m)

    # check dof = 0
    check_dof(m, fail_flag=True)

    # initialize
    m.fs.stream.initialize()

    # check solve
    results = solver.solve(m)
    assert_optimal_termination(results)

    # # check values
    assert value(m.fs.stream[0].mole_frac_phase_comp['Liq', 'Na_+']) == pytest.approx(8.845e-3,  rel=1e-3)
    assert value(m.fs.stream[0].mole_frac_phase_comp['Liq', 'Ca_2+']) == pytest.approx(1.74e-4,  rel=1e-3)
    assert value(m.fs.stream[0].mole_frac_phase_comp['Liq', 'Cl_-']) == pytest.approx(1.048e-2,  rel=1e-3)
    assert value(m.fs.stream[0].mole_frac_phase_comp['Liq', 'H2O']) == pytest.approx(0.9790,  rel=1e-3)
    assert value(m.fs.stream[0].mole_frac_phase_comp['Liq', 'Mg_2+']) == pytest.approx(1.049e-3,  rel=1e-3)
    assert value(m.fs.stream[0].mole_frac_phase_comp['Liq', 'SO4_2-']) == pytest.approx(4.07e-4,  rel=1e-3)


